/*
--------------------------------------------------------------------------------
Lawnmaker  V1.2 30.12.03
written 2003 by Uwe Glei
comments to:
cogra@cogra-flg.homelinux.net
this stuff is totally free, i hope you will find it useful.
--------------------------------------------------------------------------------
*/

/*
-----------------------------------------------------------------
 Necessary values
-----------------------------------------------------------------
Lawnmaker needs several values to distinguish the different aspects of
one patch of lawn, a pasture or whatsoever. These have to be declared before
the grasspatch itself may be invoked by LawnMaker().
To create several patches, simply reinvoke LawnMaker after changing the 
values as you see fit.
Below, each Variable is set with a deffault value, so that you may only change
what is neccessary. Here the meaning of each will be explained:
*/                

/*
All plants which are to be used are included from other inc-files.
In there could be nearly everything, POV makes possible. But keep in mind, that
LawnMaker is normaly used to place LARGE amounts of objects, so don't use
too complicated objects or be sure to have your Computer well established with
memory ;-).
To prevent to much chaos, these objects should be 1 unit high, so that scaling
may be done convienently later.
LMgrassblades.inc produces 5 differently strong curved simple grassblades as
parametric objects and 5 grassblades with cut tips, which is already enough to
produce a nice piece of lawn.
*/
#include "Lawnmaker/LMgrassblades.inc"

/*
LawnStart and LawnStop define the corners of a rectangle in the x-z-plane
in which the plants will grow. The exact shape of the overgrown area can be
controlled via a pigment-pattern. The red-value of this pattern is converted
into a scaling value for the different plants. Every type of plant gets its
own range of scaling values (see below) but black areas of the pattern will
always result in NO plants at all in those areas.
*/
#ifndef (LawnStart)		#declare LawnStart = <0, 0>;						#end
#ifndef (LawnStop)		#declare LawnStop  = <1, 1>;						#end
#ifndef (LawnHeight)	#declare LawnHeight = pigment {rgb 1}		#end

/*
The LawnStep-Values do two things:
The first component of each vector determines the average space between two
plants in that direction. The lower these values, the more dense the plants
will grow (and the more parsing- and tracing time will be necessary, so be
careful here).
The second component controls the amount of random perturbation of the
plant positions. 0 here means lined plants (i.e. as potatoes in their rows)
and a good value for a patch of lawn is to use the same value as for spacing.
*/
#ifndef (LawnStepX)		#declare LawnStepX = <0.2, 0.2>;				#end
#ifndef(LawnStepZ)		#declare LawnStepZ = <0.2, 0.2>;				#end

/*
To map the plants onto a heightfield, that heightfield is also used to determine
the vertical position of each plant.
In most cases this will be a picture, but as default, this map is set to a constant
zero, which is useful while running test renders.
The vertical scaling factor used for the heightfield is passed over with the value
LawnPosStretch and the heightmap is stored in LawnPosition.
If you are using a picture, then don't forget to rotate it into the x-z-plane
with rotate <90, 0, 0>.
*/
#ifndef (LawnPosStretch)		#declare LawnPosStretch = 10;											#end
#ifndef (LawnPosition)			#declare LawnPosition = pigment { color rgb 0 }		#end
//pigment { image_map { tga "Heightmap.tga"} rotate <90,0,0> scale 600}

/*
To control the distribution of the different plant types, another pigment-pattern
comes into use. LawnType will, in most cases, be a pattern with high turbulence
and a small scaling-value. But it also may be used to produce a flower-bed with
specific areas to use for the different plants.
How often or rare a certain type of plant may appear in a random pattern is
controlled by corresponding weight-values for each plant type (see below).
Remember that it is even possible to change the distribution of plants for
different heights if a pattern which changes with altitude is used.
*/
#ifndef (LawnType)		#declare LawnType = pigment { marble color_map {[0, rgb 0] [1, rgb 1]} turbulence 1 scale 0.01 }	#end

/*
Every single plant will be turned into a different direction and inclined more
or less to the ground. Random values for this purpose would work fine for
lawn and similar irregular growing plants, but if the goal is a field of grain
where almost all plants are bent in similar direction by the wind, a
pigment-pattern is more appropriate.
LawnDirection controls the angle each plant is turned around its vertical
axis and LawnInclination controls the bending to the ground.
By animating these patterns, it may be possible to simulate wind over a field
of wheat or even let a person leave tracks in the grass.
*/
#ifndef (LawnDirection)		#declare LawnDirection = pigment { marble color_map {[0, rgb 0] [1, rgb 1]} turbulence 1 scale 0.2 }	#end
#ifndef (LawnInclination)	#declare LawnInclination = pigment { bozo color_map {[0, rgb 0] [1, rgb 1]} turbulence 1 scale 0.2 }	#end

/*
If every single little plant has the same colour, a patch of lawn would look like
plastic and not quite natural. On the other hand, assigning one pattern to the
whole patch of grass would seem like the blades are cut out of one block of
material.
There may be a normal- and/or a finish-definition added to the plant-models in
their separate include files. Any pigment-definitions will be overriden by
Lawn-Maker. To every single plant there will be assigned one special colour.
This colour gets picked out of a pattern. So by changing that pattern, the
scattering of differently coloured blades can be controlled.
This method of colouring can be deactivated for each individual type of plant
(see below).
NOTE: This is the major change in V1.2. LawnMaker 1.0 files will work with V1.2
but may look a bit different (because the default colourmaps below have changed).
*/
#ifndef (LawnColor)
#declare LawnColor = 
array[5] {
  // fresh green pastures [0]
  pigment {
    marble turbulence 1 scale 0.5
    colour_map {
      [0.0 color rgb <0.09, 0.54, 0.03>]
      [0.5 color rgb <0.14, 0.71, 0.04>]
      [1.0 color rgb <0.22, 0.87, 0.05>]
    }
  },
  // dark natural grass [1]
  pigment {
    marble turbulence 1 scale 0.5
    colour_map {
      [0.0 color rgb <0.08, 0.20, 0.01>]
      [0.5 color rgb <0.10, 0.35, 0.03>]
      [1.0 color rgb <0.15, 0.47, 0.03>]
    }
  },
  // brown patches [2]
  pigment {
    spotted turbulence 2 scale 0.5
    colour_map {
      [0.0 color rgb <0.23, 0.37, 0.03>]
      [0.4 color rgb <0.28, 0.52, 0.04>]
      [0.6 color rgb <0.35, 0.24, 0.05>]
      [1.0 color rgb <0.55, 0.50, 0.15>]
    }
  },
  // a test colour [3]
  pigment {color rgb <1, 0, 0>},
  // a test colour [4] when defined in scene, used for the straw halms
  pigment {color rgb 0 }
} //end of array
#end

/*
LMRandom is the random seed that is used for all random values in LawnMaker.
Changing this value will result in a slightly different patch of plants with
the same characteristics though.
*/
#ifndef (LMRandom)		#declare LMRandom = 124;	#end

/*
Which plants should appear in the grasspatch is controlled with several arrays.
To prevent disorder and mistakes, you should try to place the corresponding values
in each array above each other because in this way the values for one type build
a column.
The values are:
- T: the number of plants used.
- Type Geometry:
  Here the name of the object should appear as defined before in an appropriate
  inc-file.
- Type Weight:
  This will control how many plants of that type will appear. For that purpose,
  LawnMaker adds the weights of all plants and then builds a 'plant_map' which
  assigns parts of differing length to the different plant types.
  In the example below, Blade1 will be chosen if the red-value in the pattern
  LawnType (see above) is in [0; 0.5], Values from 0.5 to 0.9 will result in Blade2
  and 0.9 to 1 will produce Blade3.
  By using this knowledge to model an appropriate pattern in LawnType, the
  area for different types of plants can be controlled precisely.
  Note: by that way the actual amount of each plant may vary from the intended
        distribution with the weight-values (i.e. a pattern which only produces
        the value 1 will result only in the plant type Blade3 over the whole patch)
- TypeSize1/2:
  These set the range of scaling-factors used for the coresponding plant-type.
  The red-value out of the LawnHeight-pattern is converted into a value between
  TypeSize1 and TypeSize2. Some geometries of some plant models are meant to be scaled
  as you like (i.e. the basic grass blades in LMgrassblades.inc) whether others
  tend to look strange if scaled to much up or down. With these values, scaling can
  be controlled for each individual plant.
  Remember, that the value 0 in the pattern will result in no plants at that point
  regardless of these two settings.
- TypeColMap:
  This selects, which of the 5 above colour-patterns (LawnColor, see above) is
  used for this plant type (the numbering of the patterns starts with 0 not with 1!!!).
  -1 used here results in no overlayed colours which means that
  the corresponding plant will totally retain its original textures (useful for all
  'non-mass-plants'.
- TypeTurn:
  The 2-component vectors contain the minimum and maximum of the angle-range
  this plant type may be turned arround its vertical axis.
- TypeIncline:
  The same as above for the inclination towards the ground.
  These angle values may be used to control which plants are more bent and turned by
  wind or other factors.
*/
/*
#declare T = 3;
#declare TypeGeometry = array[T] { Blade1,       Blade2,       Blade3  }
#declare TypeWeight   = array[T] { 5,            4,            1  }
#declare TypeSize1    = array[T] { <1,1.3,1>,    <1,1.2,1>,    <1,1,1> }
#declare TypeSize2    = array[T] { <1,1.6,1>,    <1,1.4,1>,    <1,1.2,1> }
#declare TypeColMap   = array[T] { 0,            0,            0  }      // -1 shuts of mapped coloring
#declare TypeTurn     = array[T] { <0, 360>,     <0, 360>,     <0, 360> }
#declare TypeIncline  = array[T] { <0, 0>,       <0, 0>,       <0, 0> }
*/
/*
-----------------------------------------------------------------
 macro for plant growth
-----------------------------------------------------------------
*/
#macro LawnMaker()
  // Settings for the whole lawnpatch
  //---------------------------------
  // coordinates of the lower left corner       
  #local sx = LawnStart.x;                      
  #local sz = LawnStart.y;                      
  // upper right corner 
  #local ex = LawnStop.x;
  #local ez = LawnStop.y;
  
  // heightmap of the terrain to position the blades correctly
  #local posmap = function { pigment { LawnPosition } }
  // this function controls the distribution of different grastypes
  #local grastypemap = function { pigment { LawnType } }
  // function to evaluate the heighth (the size) of the different blades
  #local grasheightmap = function { pigment { LawnHeight } }
  // distribution of the DIRECTION the bladetips point to
  #local grasdirmap = function { pigment { LawnDirection } }
  // function to determin how much the blades are bent to the ground
  #local grasincmap = function { pigment { LawnInclination } }
  // colorpatterns used to get the single plants colour
  #local grascolmap = array[5]
  {
    function { pigment { LawnColor[0] } }
    function { pigment { LawnColor[1] } }
    function { pigment { LawnColor[2] } }
    function { pigment { LawnColor[3] } }
    function { pigment { LawnColor[4] } }
  }
  
  // Calculations to distribute the plants according to the programmed
  // probabilities (Weight)
  // -----------------------------------------------------------------
  // sum of all weights
  #local ii = 0;
  #local Weights = 0;
  #while (ii<T)      
    #local Weights = Weights + TypeWeight[ii];
    #local ii = ii+1;
  #end

  // calculation of a gradient that contains differently spaced areas
  // for the different planttypes
  #local ii = 0;
  #local TWeight = array[T]
  #local Weights2 = 0;
  #declare gettype = function {
    pigment { gradient x color_map { [ 0, rgb 0]
                                     #while (ii<T)      
                                       #local Weights2 = Weights2 + TypeWeight[ii];
                                       [ Weights2/Weights, rgb <1,0,0>*ii ]
                                       #local ii = ii+1;
                                       #if (ii-T)
                                         [ Weights2/Weights, rgb <1,0,0>*ii ]
                                       #end
                                     #end
                                    }
    }
  }
  
  // the grasspatch itself
  // ---------------------
  #declare GrassPatch =
  union {
  #local R1 = seed(LMRandom);
    #local numx = sx;
    #while (numx<ex)
      #local numx = numx + LawnStepX.x;
      #local numz = sz;
    
      #while (numz < ez)    
        #local numz = numz + LawnStepZ.x;
      
        // Calculate the plants position on the heightfield
        #local xx = numx + rand(R1)*LawnStepX.y;
        #local zz = numz + rand(R1)*LawnStepZ.y;             
        // get the position in vertical direction out of the heightmap
        #local yy = posmap(xx, 0, zz).x*LawnPosStretch;
        //#debug concat ("\nyy: ", vstr(3, yy, ", ", 3, 3), ".\n")
     

        #if (grasheightmap (xx,yy,zz).x >0)  // value 0 will result in no plant at all
        
          // determine the plant type
          #local typ =  gettype(grastypemap(xx, yy, zz).x,0,0).x;//mint + int((maxt-mint)*grastypemap(xx, yy, zz).x);
                                   
          // calculate scalingfactor and turnangles
          #local groesse = TypeSize1[typ] + (TypeSize2[typ] - TypeSize1[typ])*grasheightmap (xx, yy, zz).x;  
          #local drehung = TypeTurn[typ].x + (TypeTurn[typ].y - TypeTurn[typ].x)*grasdirmap(xx, yy, zz).x;  
          #local inklination = TypeIncline[typ].x + (TypeIncline[typ].y - TypeIncline[typ].x)*grasincmap(xx, yy, zz).x;
        
          object { TypeGeometry[typ]
            // positioning and scaling of the individual plant
            scale groesse          
            rotate <0, drehung, 1*inklination>
            translate <xx, yy, zz>
          
            // colour of the plant (applied only if desired)
            // function to colour the blades
            #if (TypeColMap[typ]>-1) // only use a simple colour, if desired
              // pick the colorvalues out of the colormap for this planttype  
              #local CurTypeColMap = TypeColMap[typ]; 
              #local colpickR = grascolmap[CurTypeColMap](xx,yy,zz).x;
              #local colpickG = grascolmap[CurTypeColMap](xx,yy,zz).y;
              #local colpickB = grascolmap[CurTypeColMap](xx,yy,zz).z;
              //texture {
                pigment { color rgbt <colpickR, colpickG, colpickB> }
              //}
            #end
          }
        #end
      #end
    #end
  }  
  #undef gettype
#end

//-----------------------------------------------------------------
